#include <vector>
#include <deque>
#include "jarray.h"
#include <sstream>

using namespace std;

// stdout is global, so here we use global variables too
bool gnuplot_initialized=false;
vector< deque<double> > streams;

// left argument is scalar window size, right is the data point or several 
// (in a vector), corresponding to multiple plot lines. There must be the 
// same number of supplied data points on all invocations through the program
// lifetime.
jarray gnuplot(jarray x, jarray y) {
  assert(x.rank()==0);
  int window;
  x.get(window,0);
  if (!gnuplot_initialized)
    streams.resize(y.size());
  assert(streams.size()==y.size());
  // add new data
  jarray_of_type<double> yd(y);
  for(int i=0;i<y.size();i++)
    streams[i].push_back(yd[i]);
  // remove old data
  for(int i=0;i<y.size();i++)
    if (streams[i].size()>window)
      streams[i].pop_front();
  // rerender the frame
  ostringstream out;
  if (!gnuplot_initialized) {
    out << "plot ";
    for (int i=0; i<y.size();i++) {
      if (i>0) out << ", ";
      out << "'-'";
    };
    out << endl;
    gnuplot_initialized=true;
  } else
    out << "replot" << endl;
  // print each stream's data
  for (int i=0; i<y.size();i++) {
    for(int j=0;j<streams[i].size();j++) {
      out << streams[i][j] << endl;
    };
    out << endl << 'e' << endl;
  };
  jarray res(out.str());
  return res;
};

jarray gnuplot(jarray y) {
  jarray_of_type<jarray::I> window;
  window[0]=50;
  return gnuplot(window,y);
};
